# -*- coding: utf-8 -*-
"""
Specifies utility functions.

@author: Ivan Zubarev, ivan.zubarev@aalto.fi
"""
import os
import pickle
import warnings
import numpy as np
import tensorflow as tf
import scipy.io as sio
import mne
from mneflow import models

   

class MetaData():
    r"""
    Class containing all metadata required to run model training, prediction, 
    and evaluation. Produced by mneflow.produce_tfrecords. 
    
    Attributes
    ----------
        path : str
            A path where the output TFRecord and corresponding metadata
            files will be stored.

        data_id : str
            Filename prefix for the output files.

        input_type : str {'trials', 'continuous', 'seq', 'fconn'}
            Type of input data.

            'trials' - treats each of n inputs as an iid sample, produces dataset
            with dimensions (n, 1, t, ch)

            'seq' - treats each of n inputs as a seqence of shorter segments,
                    produces dataset with dimensions
                    (n, seq_length, segment, ch)

            'continuous' - treats inputs as a single continuous sequence,
                            produces dataset with dimensions
                            (n*(t-segment)//aug_stride, 1, segment, ch)

        target_type : str {'int', 'float'}
            Type of target variable.

            'int' - for classification,
            'float' - for regression problems.
            'signal' - regression or classification a continuous (possbily
                       multichannel) data. Requires "transform_targets" function
                       to apply to targets

        n_folds : int, optional
            Number of folds to split the data for training/validation/testing.
            One fold of the n_folds is used as a validation set.
            If test_set == 'holdout' generates one extra fold
            used as test set. Defaults to 5

        test_set : str {'holdout', 'loso', None}, optional
            Defines if a separate holdout test set is required.
            'holdout' saves 50% of the validation set
            'loso' saves the whole dataset in original order for
            leave-one-subject-out cross-validation.
            None does not leave a separate test set. Defaults to None.
            
        fs : float, optional
             Sampling frequency, required only if inputs are not mne.Epochs
    """
    
    def __init__(self):
        self.data = dict()
        self.preprocessing = dict()
        self.model_specs = dict()
        self.train_params = dict()
        self.patterns = dict()
        
    def save(self):
        """Saves the metadata to self.data['path'] + self.data['data_id']"""
        if 'path' in self.data.keys() and 'data_id' in self.data.keys():
            fname = self.data['data_id'] + '_meta.pkl'
            with open(os.path.join(self.data['path'], fname), 'wb') as f:
                pickle.dump(self, f)
            print("""Saving MetaData as {} \n
                  to {}""".format(fname, self.data['path']))
        else:
            print("""Cannot save MetaData! \n
                  Please specify meta.data['path'] and meta.data['data_id'']""")
        
        
    def restore_model(self):
        """Restored previously saved model from metadata."""
        
        if self.model_specs['scope'] == 'lfcnn':
            model = models.LFCNN(meta=self)
        elif self.model_specs['scope'] == 'varcnn':
            model = models.VARCNN(meta=self)
        elif self.model_specs['scope'] == 'fbcsp-ShallowNet':
            model = models.FBCSP_ShallowNet(meta=self)
        elif self.model_specs['scope'] == 'deep4':
            model = models.Deep4(meta=self)
        elif self.model_specs['scope'] == 'eegnet8':
            model = models.EEGNet(meta=self)
        
        model_name = "_".join([self.model_specs['scope'],
                               self.data['data_id']])
        model.km = tf.keras.models.load_model(os.path.join(self.model_specs['model_path'],
                                              model_name))
        
        #model.build()
        model.cv_patterns = self.patterns
        #TODO: set weights from self.km.weights
        #TODO: set val loss, patterns, etc, specs
        return model
    
    def update(self, data=None, preprocessing=None, train_params=None, 
               model_specs=None, patterns=None):
        """Updates metadata file"""
        if isinstance(data, dict):
            self.data.update(data)
            print("Updating: meta.data")
        if isinstance(preprocessing, dict):
            self.preprocessing.update(preprocessing)
            print("Updating: meta.preprocessing")
        if isinstance(train_params, dict):
             self.train_params.update(train_params)
             print("Updating: meta.train_params")
        if isinstance(model_specs, dict):
             self.model_specs.update(model_specs)
             print("Updating: meta.model_specs")
        if isinstance(patterns, dict):
            self.patterns.update(patterns)
            print("Updating: meta.patterns")
        #self.save()
        return
        

def _onehot(y, n_classes=False):
    """
    Transforms n-by-1 vector of class labels into n-by-n_classes array of
    one-hot encoded labels

    Parameters
    ----------
    y : array of ints
        Array of class labels

    n_classes : int
        Number of classes. If set to False (default) n_classes is set to number of
        unique labels in y


    Returns
    -------
    y_onehot : array
        array of onehot encoded labels

    """
    if not n_classes:
        """Create one-hot encoded labels."""
        n_classes = len(set(y))
    out = np.zeros((len(y), n_classes))
    for i, ii in enumerate(y):
        out[i][ii] += 1
    y_onehot = out.astype(int)
    return y_onehot


def load_meta(path, data_id=''):
    # TODO: expand functionality?
    """Load a metadata file.

    Parameters
    ----------
    path : str
        Path to TFRecord folder

    Returns
    -------
    meta : MetaData
        Metadata file

    """
    with open(path+data_id+'_meta.pkl', 'rb') as f:
        meta = pickle.load(f)
    return meta

# def load_model(model_path):
    
#     return model 




def scale_to_baseline(X, baseline=None, crop_baseline=False):
    """Perform global scaling based on a specified baseline.

    Subtracts the mean of each channel and divides by the standard deviation of
    all channels during the specified baseline interval.

    Parameters
    ----------
    X : ndarray
        Data array with dimensions [n_epochs, n_channels, time].

    baseline : tuple of int, None
        Baseline definition (in samples). If baseline is set to None (default)
        the whole epoch is used for scaling.

    crop_baseline : bool
        Whether to crop the baseline after scaling is applied. Only used if
        baseline is specified.
    Returns
    -------
    X : ndarray
        Scaled data array.

    """
    #X = X_.copy()

    if baseline is None:
        print("No baseline interval specified, scaling based on the whole epoch")
        interval = np.arange(X.shape[-1])
    elif isinstance(baseline, tuple):
        print("Scaling to interval {:.1f} - {:.1f}".format(*baseline))
        interval = np.arange(baseline[0], baseline[1])
    X0m = X[..., interval].mean(axis=2, keepdims=True)
    X0sd = X[..., interval].std(axis=(1,2), keepdims=True)

    X -= X0m
    X /= X0sd
    if crop_baseline and baseline is not None:
        X = np.delete(X, interval, axis=-1)
    #print("Scaling Done")
    return X


def _make_example(X, y, n, target_type='int'):
    """Construct a serializable example proto object from data and
    target pairs."""

    feature = {}
    feature['X'] = tf.train.Feature(
            float_list=tf.train.FloatList(value=X.flatten()))
    feature['n'] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=n.flatten()))

    if target_type == 'int':
        feature['y'] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=y.flatten()))
    elif target_type in ['float', 'signal']:
        y = y.astype(np.float32)
        feature['y'] = tf.train.Feature(
                float_list=tf.train.FloatList(value=y.flatten()))
    else:
        raise ValueError('Invalid target type.')

    # Construct the Example proto object
    example = tf.train.Example(features=tf.train.Features(feature=feature))
    return example


def _write_tfrecords(X_, y_, n_, output_file, target_type='int'):
    """Serialize and write datasets in TFRecords format.

    Parameters
    ----------
    X_ : list of ndarrays
        (Preprocessed) data matrix.
        len = `n_epochs`, shape = `(squence_length, n_timepoints, n_channels)`

    y_ : list of ndarrays
        Class labels.
        len =  `n_epochs`, shape = `y_shape`

    n_ : int
        nubmer of training examples

    output_file : str
        Name of the TFRecords file.
    """
    writer = tf.io.TFRecordWriter(output_file)

    for X, y, n in zip(X_, y_, n_):
        # print(len(X_), len(y_), X.shape, y.shape)
        X = X.astype(np.float32)
        n = n.astype(np.int64)
        # Feature contains a map of string to feature proto objects
        example = _make_example(X, y, n, target_type=target_type)
        # Serialize the example to a string
        serialized = example.SerializeToString()
        # write the serialized object to the disk
        writer.write(serialized)
    writer.close()


def _split_indices(X, y, n_folds=5):
    # TODO: check if indices are permuted
    """Generate indices for n-fold cross-validation"""
    n = X.shape[0]
    print('n:', n)
    #original_indices = np.arange(n)
    shuffle = np.random.permutation(n)
    subset_proportion = 1./float(n_folds)
    fold_size = int(subset_proportion*n)
    folds = [shuffle[i*fold_size:(i+1)*fold_size] for i in range(n_folds)]
    return folds


def _split_sets(X, y, folds, ind=-1, sample_counter=0):
    """Split the data returning a single fold specified by ind as a holdout set
        and the rest of the data as training/validation sets.

    Parameters
    ----------
    X : ndarray
        (Preprocessed) data matrix.
        shape (n_epochs, ...)

    y : ndarray
        Class labels.
        shape (n_epochs, ...)

    folds : list of arrays
        fold indices

    ind : index of the selected fold, defaults to -1

    Returns
    -------
    X_train, y_train, X_test, y_test : ndarray
        Pairs of data / targets split in Training and Validation sets.

    test_fold : np.array
        Array of indices of data samples in the held out fold


    """

    fold = folds.pop(ind) - sample_counter
    X_test = X[fold, ...]
    y_test = y[fold, ...]
    X_train = np.delete(X, fold, axis=0)
    y_train = np.delete(y, fold, axis=0)
    test_fold = fold + sample_counter
    # return X_train, np.squeeze(y_train), X_val, np.squeeze(y_val)
    return X_train, y_train, X_test, y_test, test_fold


def import_data(inp, array_keys={'X': 'X', 'y': 'y'}):
    """Import epoch data into `X, y` data/target pairs.

    Parameters
    ----------
    inp : list, mne.epochs.Epochs, str
        List of mne.epochs.Epochs or strings with filenames.
        If input is a single string or Epochs object, it is first converted
        into a list.

    array_keys : dict, optional
        Dictionary mapping {'X': 'data_matrix', 'y': 'labels'},
        where 'data_matrix' and 'labels' are names of the corresponding
        variables, if the input is paths to .mat or .npz files.
        Defaults to {'X': 'X', 'y': 'y'}

    Returns
    -------
    data, targets: ndarray
        data.shape =  [n_epochs, channels, times]

        targets.shape =  [n_epochs, y_shape]

    """
    if isinstance(inp, (mne.epochs.EpochsFIF, mne.epochs.BaseEpochs)):
        print('processing epochs')

        inp.load_data()
        data = inp.get_data()
        events = inp.events[:, 2]


    elif isinstance(inp, tuple) and len(inp) == 2:
        print('importing from tuple')
        data, events = inp

    elif isinstance(inp, str):
        # TODO: ADD CASE FOR RAW FILE
        fname = inp
        if fname[-3:] == 'fif':
            epochs = mne.epochs.read_epochs(fname, preload=True,
                                            verbose='CRITICAL')
            print(np.unique(epochs.events[:, 2]))
            events = epochs.events[:, 2]
            epochs.crop(tmin=-1., tmax=1.)
            data = epochs.get_data()


        else:
            if fname[-3:] == 'mat':
                datafile = sio.loadmat(fname)

            if fname[-3:] == 'npz':
                print('Importing from npz')
                datafile = np.load(fname)

            data = datafile[array_keys['X']]
            events = datafile[array_keys['y']]
            print('Extracting target variables from {}'
                  .format(array_keys['y']))
    else:
        print("Dataset not found")
        return None, None

    data = data.astype(np.float32)

    # Make sure that X is 3d here
    while data.ndim < 3:
        # (x, ) -> (1, 1, x)
        # (x, y) -> (1, x, y)
        data = np.expand_dims(data, 0)

    return data, events


def produce_tfrecords(inputs, 
                      path, 
                      data_id, 
                      fs=1.,
                      input_type='trials',
                      target_type='int',
                      array_keys={'X': 'X', 'y': 'y'},
                      n_folds=5,
                      scale=False,
                      scale_interval=None,
                      crop_baseline=False,
                      segment=False,
                      aug_stride=None,
                      seq_length=None,
                      picks=None,
                      overwrite=False,
                      test_set=False,
                      bp_filter=False,
                      decimate=False,
                      combine_events=None,
                      transform_targets=False,
                      scale_y=False,
                      save_as_numpy=False):

    r"""
    Produce TFRecord files from input, apply (optional) preprocessing.

    Calling this function will convert the input data into TFRecords
    format that is used to effiently store and run Tensorflow models on
    the data.


    Parameters
    ----------
    inputs : mne.Epochs, list of str, tuple of ndarrays
        Input data.

    path : str
        A path where the output TFRecord and corresponding metadata
        files will be stored.

    data_id : str
        Filename prefix for the output files.

    fs : float, optional
         Sampling frequency, required only if inputs are not mne.Epochs

    input_type : str {'trials', 'continuous', 'seq', 'fconn'}
        Type of input data.

        'trials' - treats each of n inputs as an iid sample, produces dataset
        with dimensions (n, 1, t, ch)

        'seq' - treats each of n inputs as a seqence of shorter segments,
                produces dataset with dimensions
                (n, seq_length, segment, ch)

        'continuous' - treats inputs as a single continuous sequence,
                        produces dataset with dimensions
                        (n*(t-segment)//aug_stride, 1, segment, ch)

    target_type : str {'int', 'float'}
        Type of target variable.

        'int' - for classification,
        'float' - for regression problems.
        'signal' - regression or classification a continuous (possbily
                   multichannel) data. Requires "transform_targets" function
                   to apply to targets

    n_folds : int, optional
        Number of folds to split the data for training/validation/testing.
        One fold of the n_folds is used as a validation set.
        If test_set == 'holdout' generates one extra fold
        used as test set. Defaults to 5

    test_set : str {'holdout', 'loso', None}, optional
        Defines if a separate holdout test set is required.
        'holdout' saves 50% of the validation set
        'loso' saves the whole dataset in original order for
        leave-one-subject-out cross-validation.
        None does not leave a separate test set. Defaults to None.


    segment : bool, int, optional
        If specified, splits the data into smaller segments of specified
        number of time points. Defaults to False

    aug_stride : int, optional
        Sliding window agumentation stride parameter.
        If specified, sets the stride (in time points) for 'segment'
        allowing to extract overalapping segments. Has to be <= segment.
        Only applied within each fold to prevent data leakeage. Only applied
        if 'segment' is not False. If None, then it is set equal to length of
        the 'segment' returning non-overlapping segments.
        Defaults to None.


    scale : bool, optional
        Whether to perform scaling to baseline. Defaults to False.

    scale_interval : NoneType, tuple of ints,  optional

        Baseline definition. If None (default) scaling is
        performed based on all timepoints of the epoch.
        If tuple, then baseline is data[tuple[0] : tuple[1]].
        Only used if scale == True.

    crop_baseline : bool, optional
        Whether to crop baseline specified by 'scale_interval'
        after scaling. Defaults to False.

    array_keys : dict, optional
        Dictionary mapping {'X':'data_matrix','y':'labels'},
        where 'data_matrix' and 'labels' are names of the
        corresponding variables if the input is paths to .mat or .npz
        files. Defaults to {'X':'X', 'y':'y'}

    bp_filter : bool, tuple, optional
        Band pass filter. Tuple of int or NoneType.

    decimate : False, int, optional
        Whether to decimate the input data. Defaults to False.

    combine_events : dict, optional
        Dictionary for combining or otherwise manipulating lables.
        Should contain mapping {old_label: new_label}. If provided and
        some old_labels are not specified in keys, the corresponding
        epochs are discarded.

    picks : ndarray of int, optional
        Array of channel indices to use in decoding.

    transform_targets : callable, optional
        custom function used to transform target variables

    seq_length : int, optional
        Length of segment sequence.

    overwrite : bool, optional
        Whether to overwrite the metafile if it already exists at the
        specified path.

    Returns
    -------
    meta : dict
        Metadata associated with the processed dataset. Contains all
        the information about the dataset required for further
        processing with mneflow.
        Whenever the function is called the copy of metadata is also
        saved to data_path/meta.pkl so it can be restored at any time.


    Notes
    -----
    Pre-processing functions are implemented mostly for for convenience
    when working with array inputs. When working with mne.epochs the
    use of the corresponding mne functions is preferred.

    Examples
    --------
    >>> meta = mneflow.produce_tfrecords(input_paths, \**import_opts)

    """
    assert input_type in ['trials', 'seq', 'continuous', 'fconn'], "Unknown input type."
    assert target_type in ['int', 'float', 'signal'], "Unknown target type."
    
    if not os.path.exists(path):
        os.mkdir(path)
    data_path = os.path.join(path, 'tfrecords')
    if not os.path.exists(data_path):
        os.mkdir(data_path)
    meta_fname = os.path.join(path, data_id+'_meta.pkl')
    if not os.path.exists(meta_fname) or  overwrite:
        #print("(Re)-importing data")
        train_size = 0
        test_size = 0
        val_size = 0
        folds = []
        train_paths=[]
        test_paths=[]
        
        jj = 0
        if test_set == 'holdout':
            n_folds += 1

        #meta['fs'] = fs

        if not isinstance(inputs, list):
            inputs = [inputs]
        #print("inputs:", inputs, len(inputs), type(inputs))
        if len(inputs) == 0:
            print("Cannot process Input: {} of type {}".format(inputs, type(inputs)))
            return
        for inp in inputs:
            #print("inp:", inp, len(inp), type(inp))
            

            data, events = import_data(inp, array_keys=array_keys)

            if np.any(data) == None:
                return

            else:
                if input_type == 'continuous':
                    # if input is a continuous signal ensure that target
                    # variable has shape (n_epochs, channels, time)
                    # TODO: replace with "target type?"
                    while events.ndim < 3:
                        events = np.expand_dims(events, 0)
                else:
                    # if input is trials, ensure that target variable has shape
                    # (n_trials, y_shape)
                    if events.ndim < 2:
                        events = np.expand_dims(events, -1)

                if input_type == 'trials':
                    segment_y = False
                else:
                    segment_y = True

                if input_type == 'fconn':
                    assert data.shape[1] == data.shape[2], "data.shape incompatible with fconn input type"
                    print('Input shapes: X (n, ch, ch, freq) : ', data.shape,
                          'y (n, [signal_channels], y_shape) : ', events.shape,
                          '\n',
                          'input_type : ', input_type,
                          'target_type : ', target_type,
                          'segment_y : ', segment_y)

                else:
                    print('Input shapes: X (n, ch, t) : ', data.shape,
                          'y (n, [signal_channels], y_shape) : ', events.shape,
                          '\n',
                          'input_type : ', input_type,
                          'target_type : ', target_type,
                          'segment_y : ', segment_y)

                X, Y, fold_split = preprocess(
                        data, events,
                        sample_counter=train_size,
                        input_type=input_type,
                        n_folds=n_folds,
                        scale=scale,
                        scale_interval=scale_interval,
                        crop_baseline=crop_baseline,
                        segment=segment, aug_stride=aug_stride,
                        seq_length=seq_length,
                        segment_y=segment_y)

                Y = preprocess_targets(Y, scale_y=scale_y,
                                       transform_targets=transform_targets)

                if target_type == 'int':
                    Y, n_ev, class_ratio, orig_classes = produce_labels(Y)
                    Y = _onehot(Y)
                else:
                    class_ratio = dict()
                    orig_classes = dict()


                if test_set == 'holdout':
                    X, Y, x_test, y_test, test_fold = _split_sets(X, Y,
                                                                  folds=fold_split,
                                                                  sample_counter=train_size)
                    test_size += x_test.shape[0]
                else:
                    test_fold = None
                    #TODO: remove?
#                if input_type == 'fconn':
#                    _n, meta['n_ch'], meta['n_t'], meta['n_freq'] = X.shape
#                else:
                _n, n_seq, n_t, n_ch = X.shape


                if input_type == 'seq':
                    y_shape = Y[0].shape[1:]
                else:
                    y_shape = Y[-1].shape

                n = np.arange(_n) + train_size

                train_size += _n

                # if save_as_numpy == True:
                #     train_fold = np.concatenate(folds[1:])
                #     val_fold = folds[0]
                #     np.savez(data_path+data_id,
                #              X_train=np.swapaxes(X[train_fold, ...], -2, -1),
                #              X_val=np.swapaxes(X[val_fold, ...], -2, -1),
                #              #X_test=np.swapaxes(x_test,-2, -1),
                #              y_train=Y[train_fold, ...],
                #              y_val=Y[val_fold, ...],
                #              #y_test=y_test
                #              )


                #                                                 np.min(n), np.max(n)))
                val_size += len(fold_split[0])

                print('Prepocessed sample shape:', X[0].shape)
                print('Target shape actual/metadata: ', Y[0].shape, y_shape)

                print('Saving TFRecord# {}'.format(jj))

                folds.append(fold_split)
                train_filename = ''.join([data_path, data_id, '_train_', 
                                          str(jj),
                                          '.tfrecord'])
                train_paths.append(train_filename)

                _write_tfrecords(X, Y, n, train_filename, 
                                 target_type=target_type)

                if test_set == 'loso':
                    test_size = len(Y)
                    tfrname = ''.join([data_id, '_test_', str(jj), '.tfrecord'])
                    test_filename = os.path.join(data_path, tfrname)
                                        
                    _write_tfrecords(X, Y, n, test_filename,
                                     target_type=target_type)

                elif test_set == 'holdout':
                    #meta['test_fold'].append(test_fold)
                    tfrname = ''.join([data_id, '_test_', str(jj), '.tfrecord'])
                    test_filename = os.path.join(data_path, tfrname)
                    n_test = np.arange(len(test_fold))
                    
                    _write_tfrecords(x_test, y_test, n_test, test_filename,
                                     target_type=target_type)
                    test_paths.append(test_filename)
                jj += 1
                #create and save metadata file
                
            meta_data = dict(path=path,
                             data_path=data_path, 
                             target_type=target_type,
                             input_type=input_type, 
                             data_id=data_id,
                             test_set=test_set,
                             train_paths=train_paths,
                             test_paths=test_paths,
                             folds=folds,
                             n_folds=n_folds,
                             test_fold=test_fold,
                             train_size=train_size,
                             test_size=test_size,
                             val_size=val_size,
                             n_seq=n_seq,
                             n_t=n_t,
                             n_ch=n_ch,
                             y_shape=y_shape,
                             class_ratio=class_ratio,
                             orig_classees=orig_classes,
                             fs=fs)
            
            meta_preprocessing = dict(scale=scale,
                                      scale_interval=scale_interval,
                                      crop_baseline=crop_baseline,
                                      segment=segment, aug_stride=aug_stride,
                                      seq_length=seq_length,
                                      segment_y=segment_y)
            
            meta = MetaData()
            meta.update(data=meta_data, preprocessing=meta_preprocessing)       
            
            with open(path+data_id+'_meta.pkl', 'wb') as f:
                pickle.dump(meta, f)

    elif os.path.exists(meta_fname):
        print('Metadata file found, restoring')
        meta = load_meta(path, data_id=data_id)
    else:
        print(os.path.join(path, data_id+'_meta.pkl'), "Does not exit, aborting")
        return
    return meta

def produce_labels(y, return_stats=True):
    """Produce labels array from e.g. event (unordered) trigger codes.

    Parameters
    ----------
    y : ndarray, shape (n_epochs,)
        Array of trigger codes.

    return_stats : bool
        Whether to return optional outputs.

    Returns
    -------
    inv : ndarray, shape (n_epochs)
        Ordered class labels.

    total_counts : int, optional
        Total count of events.

    class_proportions : dict, optional
        {new_class: proportion of new_class in the dataset}.

    orig_classes : dict, optional
        Mapping {new_class_label: old_class_label}.
    """
    classes, inds, inv, counts = np.unique(y,
                                           return_index=True,
                                           return_inverse=True,
                                           return_counts=True)
    total_counts = np.sum(counts)
    counts = counts/float(total_counts)
    class_proportions = {clss: cnt for clss, cnt in zip(inv[inds], counts)}
    orig_classes = {new: old for new, old in zip(inv[inds], classes)}
    if return_stats:
        return inv, total_counts, class_proportions, orig_classes
    else:
        return inv


def _combine_labels(labels, new_mapping):
    """Combine event labels.

    Parameters
    ----------
    labels : ndarray
        Label vector

    combine_dict : dict
        Mapping {new_label1: [old_label1, old_label2], ...}

    Returns
    -------
    new_labels : ndarray
        Updated label vector.

    keep_ind : ndarray
        Label indices.
    """
    assert isinstance(new_mapping, dict), "Invalid label mapping."
    # Find all possible label values
    print(labels)
    tmp = []
    for k, j in new_mapping.items():
        tmp.append(k)
        if not isinstance(j, (list, tuple)):
            # for simplicity, force all old_labels to be lists
            new_mapping[k] = [j]
        tmp.extend(new_mapping[k])

    # pick the exlusion value
    inv = np.min(tmp) - 1
    new_labels = inv*np.ones(len(labels), int)

    for new_label, old_label in new_mapping.items():
        # print(old_label, new_label)
        ind = [ii for ii, v in enumerate(labels) if v in old_label]
        new_labels[ind] = int(new_label)
    keep_ind = np.where(new_labels != inv)[0]
    #print(new_labels, keep_ind)
    return new_labels, keep_ind


def _segment(data, segment_length=200,
             seq_length=None,
             stride=None,
             input_type='trials'):
    """Split the data into fixed-length segments.

    Parameters
    ----------
    data : ndarray
        Data array of shape (n_epochs, n_channels, n_times)

    labels : ndarray
        Array of labels (n_epochs, y_shape)

    seq_length: int or None
        Length of segment sequence.

    segment_length : int or False
        Length of segment into which to split the data in time samples.

    stride : int, optional
        If specified, sets the stride (in time points) for 'segment'
        allowing to extract overalapping segments. Has to be <= segment.
        Only applied within each fold to prevent data leakeage. Only applied
        if 'segment' is not False. If None, then it is set equal to length of
        the 'segment' returning non-overlapping segments.
        Defaults to None.

    Returns
    -------
    data : ndarray
        Segmented data array of shape
        (n, [seq_length,] n_channels, segment_length)
        where n = (n_epochs//seq_length)*(n_times - segment_length + 1)//stride
        """
    x_out = []
    if input_type == 'trials':
        seq_length = 1

    if not stride:
        stride = segment_length

    for jj, xx in enumerate(data):

        n_ch, n_t = xx.shape
        last_segment_start = n_t - segment_length

        starts = np.arange(0, last_segment_start+1, stride)

        segments = [xx[..., s:s+segment_length] for s in starts]

        if input_type == 'seq':
            if not seq_length:
                seq_length = len(segments)
            seq_bins = np.arange(seq_length, len(segments)+1, seq_length)
            segments = np.split(segments, seq_bins, axis=0)[:-1]
            x_new = np.array(segments)
        else:
            x_new = np.stack(segments, axis=0)
#            if not events:
#                x_new = np.expand_dims(x_new, 1)

        x_out.append(x_new)
    if len(x_out) > 1:
        X = np.concatenate(x_out)
    else:
        X = x_out[0]
    print("Segmented as: {}".format(input_type), X.shape)
    return X


def cont_split_indices(data, events, n_folds=5, segments_per_fold=10):
    """
    Parameters
    ----------
    data : ndarray
            3d data array (n, ch, t)

    n_folds : int
             number of folds

    segments_per_fold : int
                        minimum number of different (non-contiguous)
                        data segments in each fold
    Returns
    -------
    data : ndarray
           3d data array (n, ch, t)

    events : nd.array
           labels

    folds : list of ndarrays
            indices for each fold

    """
    raw_len = data.shape[-1]
    # Define minimal duration of a single, non-overlapping data segment
    ind_samples = int(raw_len//(segments_per_fold*n_folds))
    
    segments = np.arange(0, raw_len - ind_samples + 1, ind_samples)
    data = np.concatenate([data[:, :, s: s + ind_samples] for s in segments])
    # Split continous data into non-overlapping segments
    events = np.concatenate([events[:, :, s: s + ind_samples] for s in segments])


    folds = _split_indices(data, events, n_folds=n_folds)
    return data, events, folds


def preprocess_realtime(data, decimate=False, picks=None,
                        bp_filter=False, fs=None):
    """
    Implements minimal prprocessing for convenitent real-time use.

    Parameters
    ----------
    data : np.array, (n_epochs, n_channels, n_times)
           input data array

    picks : np.array
            indices of channels to pick

    decimate : int
                decimation factor for downsampling

    bp_filter : tuple of ints
                Band-pass filter cutoff frequencies

    fs : int
         sampling frequency. Only used if bp_filter is used

    Returns
    -------
    """
    if bp_filter:
        print('Filtering')
        data = data.astype(np.float64)
        data = mne.filter.filter_data(data, fs, l_freq=bp_filter[0],
                                      h_freq=bp_filter[1],
                                      method='iir', verbose=False)

    if isinstance(picks, np.ndarray):
        data = data[:, picks, :]

    if decimate:
        print("Decimating")
        data = data[..., ::decimate]
    return data


def preprocess(data, events, sample_counter,
               input_type='trials', n_folds=5,
               scale=False, scale_interval=None, crop_baseline=False,
               segment=False, aug_stride=None,
               seq_length=None,
               segment_y=False):
    """Preprocess input data. Applies scaling, segmenting/augmentation,
     and defines the split into training/validation folds.

    Parameters
    ----------
    data : np.array, (n_epochs, n_channels, n_times)
           input data array

    events : np.array
            input array of target variables (n_epochs, ...)

    input_type : str {'trials', 'continuous'}
            See produce_tfrecords.

    n_folds : int
            Number of folds defining the train/validation/test split.

    sample_counter : int
            Number of traning examples in the dataset

    scale : bool, optional
        Whether to perform scaling to baseline. Defaults to False.

    scale_interval : NoneType, tuple of ints or floats,  optional
        Baseline definition. If None (default) scaling is
        performed based on all timepoints of the epoch.
        If tuple, than baseline is data[tuple[0] : tuple[1]].
        Only used if scale == True.

    crop_baseline : bool, optional
        Whether to crop baseline specified by 'scale_interval'
        after scaling (defaults to False).

    segment : bool, int, optional
        If specified, splits the data into smaller segments of specified
        number of time points. Defaults to False

    aug_stride : int, optional
        If specified, sets the stride (in time points) for 'segment'
        allowing to extract overalapping segments. Has to be <= segment.
        Only applied within each fold to prevent data leakeage. Only applied
        if 'segment' is not False. If None, then it is set equal to length of
        the 'segment' returning non-overlapping segments.
        Defaults to None.

    seq_length: int or None
        Length of segment sequence.

    segment_y : bool
        whether to segment target variable in the same way as data. Only used
        if segment != False


    Returns
    -------
    X: np.array
        Data array of dimensions [n_epochs, n_seq, n_t, n_ch]

    Y : np.array
        Label arrays of dimensions [n_epochs, *(y_shape)]

    folds : list of np.arrays

    """
    print("Preprocessing:")

    # TODO: remove scale_y and transform targets?

    if scale:
        data = scale_to_baseline(data, baseline=scale_interval,
                                 crop_baseline=crop_baseline)

    #define folds
    if input_type  == 'continuous':
        data, events, folds = cont_split_indices(data, events,
                                                 n_folds=5,
                                                 segments_per_fold=10)
        shuffle = np.random.permutation(np.arange(events.shape[0]))
        data = data[shuffle]
        events = events[shuffle]
        print("Continuous events: ", events.shape)

    else:
        shuffle = np.random.permutation(np.arange(events.shape[0]))
        data = data[shuffle]
        events = events[shuffle]
        folds = _split_indices(data, events, n_folds=n_folds)

    print("Splitting into: {} folds x {}".format(len(folds), len(folds[0])))

    if segment:
        print("Segmenting")
        X = []
        Y = []
        segmented_folds = []
        jj = 0
        for fold in folds:
            #print(data[fold, ...].shape)
            x = _segment(data[fold, ...], segment_length=segment,
                         stride=aug_stride, input_type=input_type,
                         seq_length=seq_length)

            nsegments = x.shape[0]

            # if segment_y -> segment, else-> replicate
            if segment_y:
                y = _segment(events[fold, ...], segment_length=segment,
                             stride=aug_stride, input_type=input_type,
                             seq_length=seq_length)
            else:
                print("Replicating labels for segmented data")
                y = np.repeat(events[fold, ...], nsegments//len(fold), axis=0)

            if x.ndim == 3:
                x = np.expand_dims(x, 1)
            X.append(x)
            Y.append(y)
            segmented_folds.append(np.arange(jj, jj + nsegments) + sample_counter)
            jj += nsegments
        X = np.concatenate(X, axis=0)
        Y = np.concatenate(Y, axis=0)

        folds = segmented_folds
    else:
        # If not segmented add a singleton "n_seq" dminesion to X
        if data.ndim == 3:
            X = np.expand_dims(data, 1)
        elif data.ndim == 4:
            X = data

        Y = events
        folds = [f + sample_counter for f in folds]
    # Finally cast X into shape [n_epochs, n_seq, n_times, n_channels]
    if input_type != 'fconn':
        X = np.swapaxes(X, -2, -1)

    print('Preprocessed:', X.shape, Y.shape,
          'folds:', len(folds), 'x', len(folds[0]))
    assert X.shape[0] == Y.shape[0], "n_epochs in X ({}) does not match n_epochs in Y ({})".format(X.shape[0], Y.shape[0])

    return X, Y, folds

def preprocess_targets(y, scale_y=False, transform_targets=None):

    if callable(transform_targets):
        y = transform_targets(y)

    if scale_y:
            y -= y.mean(axis=0, keepdims=True)
            y /= y.std(axis=0, keepdims=True)
    print('Preprocessed targets: ', y.shape)

    return y


def regression_metrics(y_true, y_pred):
    y_shape = y_true.shape[-1]

    cc = np.diag(np.corrcoef(y_true.T, y_pred.T)[:y_shape,-y_shape:])
    r2 =  r2_score(y_true, y_pred)
    cs = cosine_similarity(y_true, y_pred)
    bias = np.mean(y_true, axis=0) - np.mean(y_pred, axis=0)
    #ve = pve(y_true, y_pred)
    return dict(cc=cc, r2=r2, cs=cs, bias=bias)

def cosine_similarity(y_true, y_pred):
    # y_true -= y_true.mean()
    # y_pred -= y_pred.mean()

    return np.dot(y_pred.T, y_true) / (np.sqrt(np.sum(y_pred**2,axis=0)) * np.sqrt(np.sum(y_true**2, axis=0)))

def pve(y_true, y_pred):
    y_true -= y_true.mean(axis=0)
    y_pred -= y_pred.mean(axis=0)
    return np.dot(y_pred.T, y_true) / np.sum(y_pred**2, axis=0)

def r2_score(y_true, y_pred):
    res = np.sum((y_true - y_pred)**2, axis=0)
    tot = np.sum((y_true - np.mean(y_true, axis=0, keepdims=True))**2, axis=0)
    return 1 - res/tot




# -*- coding: utf-8 -*-
"""
Defines mneflow.layers for mneflow.models.

@author: Ivan Zubarev, ivan.zubarev@aalto.fi
"""
#TODO: keras compatible layers
#TODO: pooling layer

#import functools
import tensorflow as tf
from tensorflow.keras.initializers import Constant
from tensorflow.keras.activations import relu
from tensorflow.keras import constraints as k_con, regularizers as k_reg
# import tensorflow.compat.v1 as tf
# tf.disable_v2_behavior()
import numpy as np

bias_const = 0.1
bias_traiable = True

class BaseLayer(tf.keras.layers.Layer):
    def __init__(self, size, nonlin, specs, **args):
        super(BaseLayer, self).__init__(**args)
        self.size = size
        self.nonlin = nonlin
        self.specs = specs
        # self.specs.setdefault("l1_lambda", 0.)
        # self.specs.setdefault("l2_lambda", 0.)
        # self.specs.setdefault("l1_scope", [])
        # self.specs.setdefault("l2_scope", [])
        # self.specs.setdefault("maxnorm_scope", [])

    def _set_regularizer(self):
        if self.scope in self.specs['l1_scope'] or 'weights' in self.specs['l1_scope']:
            reg = k_reg.l1(self.specs['l1_lambda'])
            print('Setting reg for {}, to l1'.format(self.scope))
        elif self.scope in self.specs['l2_scope'] or 'weights' in self.specs['l2_scope']:
            reg = k_reg.l2(self.specs['l2_lambda'])
            print('Setting reg for {}, to l2'.format(self.scope))
        else:
            reg = None
        return reg

    def _set_constraints(self, axis=0):
        if self.scope in self.specs['unitnorm_scope']:
            constr = k_con.UnitNorm(axis=axis)
            print('Setting constraint for {}, to UnitNorm'.format(self.scope))
        else:
            constr = None
        return constr


class Dense(BaseLayer, tf.keras.layers.Layer):


    """
    Fully-connected layer
    """
    def __init__(self, scope="fc", size=None, nonlin=tf.identity, specs={},
                 **args):
        self.scope = scope
        super(Dense, self).__init__(size=size, nonlin=nonlin, specs=specs,
             **args)
        self.constraint = self._set_constraints()
        self.reg = self._set_regularizer()

    # def get_config(self):
    #     config = self.get_config()
    #     config.update({'scope': self.scope, 'size': self.size,
    #                    'nonlin': self.nonlin})
        # return config

    def build(self, input_shape):
        super(Dense, self).build(input_shape)
        # print(input_shape)
        self.flatsize = np.prod(input_shape[1:])
        #print(self.scope, ':::', )

        self.w = self.add_weight(shape=[self.flatsize, self.size],
                                 initializer='he_uniform',
                                 regularizer=self.reg,
                                 constraint=self.constraint,
                                 trainable=True,
                                 name='fc_weights',
                                 dtype=tf.float32)

        self.b = self.add_weight(shape=[self.size],
                                 initializer=Constant(bias_const),
                                 regularizer=None,
                                 trainable=bias_traiable,
                                 name='fc_bias',
                                 dtype=tf.float32)

        print("Built: {} input: {}".format(self.scope, input_shape))


    def call(self, x, training=None):
        """Dense layer currying, to apply layer to any input tensor `x`"""
        while True:
            with tf.name_scope(self.scope):
                if len(x.shape) > 2:  # flatten if input is not 2d array
                    x = tf.reshape(x, [-1, self.flatsize])
                tmp = tf.matmul(x, self.w) + self.b
                tmp = self.nonlin(tmp, name='out')
                #print(self.scope, ": output :", tmp.shape)
                return tmp

# class LFTConvTranspose1(tf.keras.layers.Layer):
#     def __init__(self, kernel_size, stride, output_padding, filters=None, **kwargs):
#         super(LFTConvTranspose1, self).__init__(**kwargs)
#         self.scope='lft_trans'
#         self.kernel_size = kernel_size #int
#         self.stride = stride #int
#         self.output_padding = output_padding #int
#         self.filters = filters
#         #self.use_bias = use_bias #bool
#         #???
#         self.input_ax_shape = 0
#         self.lambdas = []
#         #self.nm = name

#     def deconv_length(self, input_size, stride_size, kernel_size, output_padding=0):

#         #simple 1 dimentional case
#         # Get the dilated kernel size
#         #kernel_size = kernel_size + (kernel_size - 1) * (dilation - 1)

#         # Infer length if output padding is None, else compute the exact length
#         dim_size = input_size * stride_size - kernel_size + output_padding
#         return dim_size

#     def build(self, input_shape):
#         self.n_latent = input_shape[-1]
#         self.input_ax_shape = input_shape[-2]

#         if self.filters is not None:
#             self.trainable = False
#             self.filters = tf.transpose(self.filters, [1,2,3,0])
#             print("Using Pre-determined filters for inverse convolution")
#         else:
#             print("Using Trainable filters for inverse convolution")
#             self.filters = self.add_weight(name=self.scope + "enc_kernel",
#                                initializer='he_uniform',
#                                shape=(self.kernel_size,self.n_latent, 1, 1),
#                                trainable=True)
#         print("Filters: ", self.filters.shape)


#         #self.lambdas = tf.stack(self.lambdas,axis = 0)

#         #self.input_shape = input_shape
#         self.output_length = self.deconv_length(self.input_ax_shape,
#                                            self.stride,
#                                            self.kernel_size,
#                                            self.output_padding)
#         self.out_shape = tf.TensorShape([1, input_shape[1],
#                                          self.output_length, self.n_latent,
#                                          ])
#         #self.out_shape[-2] += (self.deconv_length - self.input_ax_shape)

#         super(LFTConvTranspose1, self).build(input_shape)
#         print("Built Enc deconv:", input_shape, "->", self.out_shape)

#     #@tf.function
#     def call(self, inputs):

#         out = tf.nn.conv2d_transpose(inputs, self.filters,
#                                      output_shape=self.out_shape,
#                                      strides=(self.stride, 1),
#                                      data_format='NCHW',
#                                      padding='SAME')

#         #out_trans = tf.transpose(out, perm=[0,3,1,2])
#         #print(out_trans.shape)

#         return out


# class LFTConvTranspose(tf.keras.layers.Layer):
#     def __init__(self, kernel_size, stride, output_padding, filters=None, **kwargs):
#         super(LFTConvTranspose, self).__init__(**kwargs)
#         self.scope='lft_trans'
#         self.kernel_size = kernel_size #int
#         self.stride = stride #int
#         self.output_padding = output_padding #int
#         self.filters = filters
#         #self.use_bias = use_bias #bool
#         #???
#         self.input_ax_shape = 0
#         self.lambdas = []
#         #self.nm = name

#     def deconv_length(self, input_size, stride_size, kernel_size, output_padding=0):

#         #simple 1 dimentional case
#         # Get the dilated kernel size
#         #kernel_size = kernel_size + (kernel_size - 1) * (dilation - 1)

#         # Infer length if output padding is None, else compute the exact length
#         dim_size = input_size * stride_size - kernel_size + output_padding
#         return dim_size

#     def build(self, input_shape):
#         self.n_latent = input_shape[-1]
#         self.input_ax_shape = input_shape[-2]

#         if self.filters is not None:
#             self.trainable = False
#             self.lambdas = tf.split(self.filters, self.n_latent, axis=-2)
#             print("Using Pre-determined filters for inverse convolution")
#         else:
#             print("Using Trainable filters for inverse convolution")
#             for i in range(input_shape[-1]):
#                self.lambdas.append(self.add_weight(name = self.scope + "_k" + str(i),
#                                                    initializer='he_uniform',
#                                                    shape=(1,self.kernel_size,1,1),
#                                                    trainable=True))


#         #self.lambdas = tf.stack(self.lambdas,axis = 0)

#         #self.input_shape = input_shape
#         self.output_length = self.deconv_length(self.input_ax_shape,
#                                            self.stride,
#                                            self.kernel_size,
#                                            self.output_padding)
#         self.out_shape = tf.TensorShape([1, input_shape[1],
#                                          self.output_length, 1])
#         #self.out_shape[-2] += (self.deconv_length - self.input_ax_shape)
#         print(self.out_shape)
#         super(LFTConvTranspose, self).build(input_shape)

#     @tf.function
#     def call(self, inputs):

#         inputs_channel_wise =   tf.split(inputs, self.n_latent, axis=-1)

#         #TODO: define strides and padding
#         # channel_wise_conv = tf.map_fn(lambda x:tf.nn.conv2d_transpose(input=x[0],
#         #                                                               filters=x[1],
#         #                                                               output_shape=out_shape,
#         #                                                               strides=(1, self.strides)),
#         #                               (inputs_channel_wise,self.lambdas),
#         #                               fn_output_signature=tf.float32)

#         # channel_wise_conv = tf.transpose(tf.squeeze(channel_wise_conv,axis = -1),[0,2,3,1])
#         channel_wise_conv = tf.concat([tf.nn.conv2d_transpose(inp, filt,
#                                           output_shape=self.out_shape,
#                                           strides=(1, self.stride),
#                                           padding='SAME')
#                    for inp, filt in zip(inputs_channel_wise, self.lambdas)],
#                   axis=-1)
#         print(channel_wise_conv.shape)
#         return channel_wise_conv

class DeMixing(BaseLayer):
    """
    Spatial demixing Layerю
    """
    def __init__(self, scope="dmx", size=None, nonlin=tf.identity, axis=-1,
                 specs={},  **args):
        self.scope = scope
        self.axis = axis
        super(DeMixing, self).__init__(size=size, nonlin=nonlin, specs=specs,
             **args)

    # def get_config(self):
    #     config = super(DeMixing, self).get_config()
    #     config.update({'scope': self.scope, 'size': self.size,
    #                    'nonlin': self.nonlin, 'axis': self.axis})
    #     return config

    def build(self, input_shape):

        super(DeMixing, self).build(input_shape)
        self.constraint = self._set_constraints(axis=0)
        self.reg = self._set_regularizer()

        self.w = self.add_weight(
                shape=(input_shape[self.axis], self.size),
                initializer='he_uniform',
                regularizer=self.reg,
                constraint = self.constraint,
                trainable=True,
                name='dmx_weights',
                dtype=tf.float32)

        self.b_in = self.add_weight(shape=([self.size]),
                                    initializer=Constant(bias_const),
                                    regularizer=None,
                                    trainable=bias_traiable,
                                    name='bias',
                                    dtype=tf.float32)
        print("Built: {} input: {}".format(self.scope, input_shape))

    #@tf.function
    def call(self, x, training=None):
        while True:
            with tf.name_scope(self.scope):
                try:
                    demix = tf.tensordot(x, self.w, axes=[[self.axis], [0]],
                                         name='dmx')
                    demix = self.nonlin(demix + self.b_in)
                    #print(self.scope, ": output :", demix.shape)
                    return demix
                except(AttributeError):
                    input_shape = x.shape
                    self.build(input_shape)
                    #print(self.scope, 'building from call')

class InvCov(BaseLayer):
    """
    Spatial demixing Layer
    """
    def __init__(self, scope="InvCov", size=None, nonlin=tf.identity, axis=-1,
                 specs={},  **args):
        self.scope = scope
        self.ax = axis
        self.axis = np.delete(np.arange(4), self.ax)[1:]
        print("axis: ", self.axis)
        super(InvCov, self).__init__(size=size, nonlin=nonlin, specs=specs,
             **args)

    # def get_config(self):
    #     config = super(InvCov, self).get_config()
    #     config.update({'scope': self.scope, 'size': self.size,
    #                    'nonlin': self.nonlin, 'axis': self.axis})
    #     return config

    def build(self, input_shape):

        super(InvCov, self).build(input_shape)
        self.constraint = self._set_constraints()
        self.reg = self._set_regularizer()
        sc = np.sum([input_shape[d] for d in self.axis])
        self.scaler = tf.constant(sc)
#        self.w = self.add_weight(
#                shape=(input_shape[self.axis], self.size),
#                initializer='he_uniform',
#                regularizer=self.reg,
#                constraint = self.constraint,
#                trainable=True,
#                name='dmx_weights',
#                dtype=tf.float32)
#
#        self.b_in = self.add_weight(shape=([self.size]),
#                                    initializer=Constant(0.1),
#                                    regularizer=None,
#                                    trainable=True,
#                                    name='bias',
#                                    dtype=tf.float32)
        print("Built: {} input: {}".format(self.scope, input_shape))

    #@tf.function
    def call(self, x, training=None):
        while True:
            with tf.name_scope(self.scope):
                try:

                    _cov = tf.tensordot(x, x, axes=[[self.axis], [self.axis]],
                                         name='_cov')
                    print("_cov", _cov.shape)

                    cov =  tf.reduce_mean(tf.divide(_cov, self.scaler),axis=0)
                    print("det:", tf.det(cov))
                    invcov = tf.linalg.pinv(cov)
                    out = tf.tensordot(invcov, x, axes = [0, self.ax])
                    print(self.scope, ": output :", out.shape)
                    return out
                except(AttributeError):
                    input_shape = x.shape
                    self.build(input_shape)
                    #print(self.scope, 'building from call')


class SquareSymm(BaseLayer):
    """
    Spatial demixing Layerю
    """
    def __init__(self, scope="ssym", size=None, nonlin=tf.identity, axis=1,
                 specs={},  **args):
        self.scope = scope
        self.axis = axis
        super(SquareSymm, self).__init__(size=size, nonlin=nonlin, specs=specs,
             **args)


    def build(self, input_shape):

        super(SquareSymm, self).build(input_shape)
        self.constraint = self._set_constraints(axis=0)
        self.reg = self._set_regularizer()

        self.w = self.add_weight(
                shape=(input_shape[self.axis], self.size),
                initializer='he_uniform',
                regularizer=self.reg,
                constraint = self.constraint,
                trainable=True,
                name='ssym_weights',
                dtype=tf.float32)

        self.b_in = self.add_weight(shape=([self.size]),
                                    initializer=Constant(0.1),
                                    regularizer=None,
                                    trainable=True,
                                    name='bias',
                                    dtype=tf.float32)
        print("Built: {} input: {}".format(self.scope, input_shape))

    #@tf.function
    def call(self, x, training=None):
        while True:
            with tf.name_scope(self.scope):
                try:
                    d1 = tf.tensordot(x, self.w, axes=[[1], [0]],
                                         name='smx') #output
                    d2 = tf.tensordot(d1, self.w, axes=[[1], [0]],
                                         name='smx')

                    demix = self.nonlin(d2 + self.b_in)
                    #demix = tf.transpose(demix. [0, 2, 3, 1])
                    #print(self.scope, ": output :", demix.shape)
                    return demix
                except(AttributeError):
                    input_shape = x.shape
                    self.build(input_shape)
                    #print(self.scope, 'building from call')


class LFTConv(BaseLayer):
    """
    Stackable temporal convolutional layer, interpreatble (LF)
    """
    def __init__(self, scope="tconv", size=32,  nonlin=tf.nn.relu,
                 filter_length=7, pooling=2, padding='SAME', specs={},
                 **args):
        self.scope = scope
        super(LFTConv, self).__init__(size=size, nonlin=nonlin, specs=specs,
             **args)
        self.size = size
        self.filter_length = filter_length
        self.padding = padding

    # def get_config(self):

    #     config = super(LFTConv, self).get_config()
    #     config.update({'scope': self.scope,
    #                    'filter_length': self.filter_length,
    #                    'nonlin': self.nonlin, 'padding': self.padding})
    #    return config

    def build(self, input_shape):
        super(LFTConv, self).build(input_shape)
        self.constraint = self._set_constraints(axis=1)
        self.reg = self._set_regularizer()
        shape = [1, self.filter_length, input_shape[-1], 1]
        self.filters = self.add_weight(shape=shape,
                                       initializer='he_uniform',
                                       regularizer=self.reg,
                                       constraint=self.constraint,
                                       trainable=True,
                                       name='tconv_weights',
                                       dtype=tf.float32)

        self.b = self.add_weight(shape=([input_shape[-1]]),
                                 initializer=Constant(bias_const),
                                 regularizer=None,
                                 trainable=bias_traiable,
                                 name='bias',
                                 dtype=tf.float32)
        print("Built: {} input: {}".format(self.scope, input_shape))

    #@tf.function
    def call(self, x, training=None):
        while True:
            with tf.name_scope(self.scope):
                try:
                    conv = tf.nn.depthwise_conv2d(x,
                                                  self.filters,
                                                  padding=self.padding,
                                                  strides=[1, 1, 1, 1],
                                                  data_format='NHWC')
                    conv = self.nonlin(conv + self.b)

                    #print(self.scope, ": output :", conv.shape)
                    return conv
                except(AttributeError):
                    input_shape = x.shape
                    self.build(input_shape)


class VARConv(BaseLayer):
    """
    Stackable temporal convolutional layer, interpreatble (LF)
    """
    def __init__(self, scope="tconv", size=32,  nonlin=tf.nn.relu,
                 filter_length=7, pooling=2, padding='SAME', specs={},
                 **args):
        self.scope = scope
        super(VARConv, self).__init__(size=size, nonlin=nonlin, specs=specs,
             **args)
        self.size = size
        self.filter_length = filter_length
        self.padding = padding


    # def get_config(self):

    #     config = super(VARConv, self).get_config()
    #     config.update({'scope': self.scope,
    #                    'filter_length': self.filter_length,
    #                    'nonlin': self.nonlin, 'padding': self.padding})
    #    return config

    def build(self, input_shape):
        print("input_shape:", input_shape)
        super(VARConv, self).build(input_shape)

        self.constraint = self._set_constraints()
        self.reg = self._set_regularizer()
        shape = [1, self.filter_length, input_shape[-1], self.size]
        self.filters = self.add_weight(shape=shape,
                                       initializer='he_uniform',
                                       regularizer=self.reg,
                                       constraint=self.constraint,
                                       trainable=True,
                                       name='tconv_weights',
                                       dtype=tf.float32)

        self.b = self.add_weight(shape=([input_shape[-1]]),
                                 initializer=Constant(bias_const),
                                 regularizer=None,
                                 trainable=bias_traiable,
                                 name='bias',
                                 dtype=tf.float32)
        print("Built: {} input: {}".format(self.scope, input_shape))

    #@tf.function
    def call(self, x, training=None):
        while True:
            with tf.name_scope(self.scope):
                try:
                    conv = tf.nn.conv2d(x, self.filters,
                                        padding=self.padding,
                                        strides=[1, 1, 1, 1],
                                        data_format='NHWC')

                    conv = self.nonlin(conv + self.b)
                    #print(self.scope, ": output :", conv.shape)
                    return conv
                except(AttributeError):
                    input_shape = x.shape
                    self.build(input_shape)
                    #print(self.scope, 'building from call')


class TempPooling(BaseLayer):
    def __init__(self, scope="pool", stride=2, pooling=2, specs={},
                 padding='SAME', pool_type='max', **args):
        self.scope = '_'.join([pool_type, scope])
        super(TempPooling, self).__init__(size=None, nonlin=None, specs=specs,
                                          **args)
        self.strides = [1, 1, stride,  1]
        self.kernel = [1, 1, pooling,  1]

        self.padding = padding
        self.pool_type = pool_type

    #@tf.function
    def call(self, x):
        if self.pool_type == 'avg':
            pooled = tf.nn.avg_pool2d(
                                x,
                                ksize=self.kernel,
                                strides=self.strides,
                                padding=self.padding,
                                data_format='NHWC')
        else:
            pooled = tf.nn.max_pool2d(
                                x,
                                ksize=self.kernel,
                                strides=self.strides,
                                padding=self.padding,
                                data_format='NHWC')
        #print(self.scope, ": output :", pooled.shape)
        return pooled

    def build(self, input_shape):
        super(TempPooling, self).build(input_shape)
        self.built = True


    # def get_config(self):
    #     config = super(TempPooling, self).get_config()
    #     config.update({'scope': self.scope,
    #                    'pool_type': self.pool_type,
    #                    'stride': self.strides, 'pooling': self.pooling,
    #                    'padding': self.padding})
    #     return config



class LSTM(tf.keras.layers.LSTM):
    def __init__(self, scope="lstm", size=32, nonlin='tanh', dropout=0.0,
                 recurrent_activation='tanh', recurrent_dropout=0.0,
                 use_bias=True, unit_forget_bias=True,
                 kernel_regularizer=None, bias_regularizer=None,
                 return_sequences=True, stateful=False, unroll=False, **args):
        super(LSTM, self).__init__(name=scope,
                                     units=size,
                                     activation=nonlin,
                                     dropout=dropout,
                                     recurrent_activation=recurrent_activation,
                                     recurrent_dropout=recurrent_dropout,
                                     use_bias=use_bias,
                                     unit_forget_bias=unit_forget_bias,
                                     kernel_regularizer=kernel_regularizer,
                                     # kernel_initializer='glorot_uniform',
                                     # recurrent_initializer='orthogonal',
                                     bias_regularizer=bias_regularizer,
                                     return_sequences=return_sequences,
                                     stateful=stateful,
                                     unroll=unroll,
                                     **args)
        self.scope = scope
        self.size = size
        self.nonlin = nonlin
        print(self.scope, 'init : OK')

    # def get_config(self):
    #     config = super(LSTM, self).get_config()
    #     config.update({'scope': self.scope, 'size': self.size,
    #                    'nonlin': self.nonlin})
    #     return config

    def build(self, input_shape):
        # print(self.scope, 'build : OK')
        super(LSTM, self).build(input_shape)

    @tf.function
    def call(self, inputs, mask=None, training=None, initial_state=None):
        # print(self.scope, inputs.shape)
        return super(LSTM, self).call(inputs, mask=mask, training=training,
                                        initial_state=initial_state)




if __name__ == '__main__':
    print('Reloaded')




import mneflow
from . import models
from . import layers
from . import utils
from . import data
from .utils import produce_tfrecords, load_meta
from .data import Dataset

mneflow.__version__ = '0.5.1'




#! /usr/bin/env python
from setuptools import setup
import codecs
import os

if __name__ == '__main__':
    if os.path.exists('MANIFEST'):
        os.remove('MANIFEST')

    setup(name='mneflow',
        maintainer='Ivan Zubarev',
        maintainer_email='ivan.zubarev@aalto.fi',
        description='Neural networks for MEG and EEG data',
        license='BSD-3',
        url='https://github.com/zubara/mneflow',
        version='0.5.1',
        download_url='https://github.com/zubara/mneflow/archive/master.zip',
        #long_description=codecs.open('./docs/intro.rst', encoding='utf8').read(),
        long_description_content_type="text/x-rst",
        classifiers=['Intended Audience :: Science/Research',
                     'Intended Audience :: Developers',
                     'License :: OSI Approved',
                     'Programming Language :: Python',
                     'Topic :: Software Development',
                     'Topic :: Scientific/Engineering',
                     'Operating System :: Microsoft :: Windows',
                     'Operating System :: POSIX',
                     'Operating System :: Unix',
                     'Operating System :: MacOS'],
        platforms='any',
        packages=['mneflow'],
        install_requires=['numpy', 'scipy', 'mne', 'tensorflow', 'matplotlib'])
